import warnings
import numpy as np
# from eral.time_series_helpers import _remove_nan_padding_single, _var_nan_sum
# from eral.time_series_alignment import sync_n_series_to_prototype, get_score_and_shift
# from .batch_cluster import BatchCluster as Cluster
from seral.seral import Cluster

from collections.abc import Generator

class EvolvingTimeSeriesClustering:  

    clusters: list[Cluster]
    outlier_clusters: list[Cluster]
    threshold: float
    exclusion_zone: float
    new_min_length: float
    alpha: float
    outlier_percentage: float
    verbose: bool

    def __init__(self, threshold: float|None = None, exclusion_zone: float = 0.2, new_min_length: float = 0.8, alpha: float = 0.1, outlier_percentage: float = 0.01, verbose: bool = True):
        """ Initialize the evolving time series clustering

        Choosing :param:`alpha`: 1.0 --> only use region where all samples were used, results in a short prototype, 
        0.0 --> use all samples, results in a long prototype

        :param threshold: Maximum distance to the closest cluster for the sample to be added to the cluster
        :param exclusion_zone: Exclusion zone for the alignment of the new sample to the prototype
        :param new_min_length: Minimum length of the new prototype as a proportion of the old prototype
        :param alpha: Threshold for PCV: required proportion of samples to be used in conditioning the prototype.
        """

        self.clusters = []
        self.outlier_clusters = []
        self.exclusion_zone = exclusion_zone
        self.threshold = threshold
        self.min_new_length = new_min_length
        self.alpha = alpha
        self.max_outlier_points = 0
        self.cluster_generator = self.get_cluster_id()
        self.outlier_percentage = outlier_percentage
        self.verbose = verbose

        super().__init__()

    def _get_threshold_curve(self, data: list[np.ndarray], expected_number_of_clusters: int):
        D = []
        for i, x in enumerate(data):
            d = []
            for j, y in enumerate(data):
                if i == j:
                    continue
                distance = Cluster.distance(x, y, self.exclusion_zone)
                d.append(distance)
            # D.append(np.min(d))
            D.append(np.percentile(d,100/(2*expected_number_of_clusters)))
        return D

    def estimate_threshold(self, data: list[np.ndarray], expected_number_of_clusters: int, percentile: float = 0.95) -> float:
        """ Estimate the threshold for the clustering

        The threshold is estimated as the `percentile`-th percentile of the distances of the samples to their closest neighbors.

        Configures the self.threshold attribute.
        
        :param data: List of samples to be used for the estimation
        :param percentile: Percentile of the distances to be used as the threshold
        :return: Estimated threshold

        """
        if percentile < 0 or percentile > 100:
            raise ValueError("Percentile must be between 0 and 100")

        D = self._get_threshold_curve(data, expected_number_of_clusters)
        
        if 0 <= percentile <= 1:
            percentile = percentile*100
        
        self.threshold = np.percentile(D, percentile)
        return self.threshold

    def fit_predict(self, data: list[np.ndarray]) -> list[int]:
        """ Fit the data to the clusters and return the cluster indices"""
        y_pred = []
        for i, sample in enumerate(data):
            y = self.fit_point(sample-np.mean(sample))
            y_pred.append(y)

            merge_success, merge_ids = self.try_merge_clusters(merge_threshold=self.threshold, id=y)
            if merge_success:
                new_id, old_id = merge_ids
                if self.verbose:
                    print(f"Merged {old_id} into {new_id}")
                y_pred = [new_id if x==old_id else x for x in y_pred]

        return y_pred

    def get_cluster_id(self) ->  Generator[list[int], None, None]:
        """ Generator for cluster ids

        :return: Generator for cluster ids
        """
        cluster_id = 0
        while True:
            yield cluster_id
            cluster_id += 1

    def try_promote_outlier(self, id: int | None = None) -> bool:
        """ Try to promote outlier clusters to regular clusters.

        Only promotes a single cluster at a time.
        
        If no id is provided, the outlier cluster with the most points is tested for promotion. If an id is provided, the outlier
        cluster with that id is tested for promotion.

        :param id: Id of the outlier cluster to be tested for promotion
        :return: True if a cluster was promoted, False otherwise
        """
        if id is None:
            sorted_clusters = sorted(self.outlier_clusters, key=lambda x: x.number_of_points, reverse=True)
            cluster_under_test = sorted_clusters[0]
        else:
            cluster_under_test = [cluster for cluster in self.outlier_clusters if cluster.id == id][0]
        
        if cluster_under_test.number_of_points > self.max_outlier_points:
            self.clusters.append(cluster_under_test)
            self.outlier_clusters.remove(cluster_under_test)
            if self.verbose:
                print(f"Outlier cluster {cluster_under_test.id} was promoted to a regular cluster. Now {len(self.clusters)} regular clusters and {len(self.outlier_clusters)} outlier clusters.")
            return True
        return False


    def try_demote_cluster(self) -> bool:
        """ Try to demote regular clusters to outlier clusters.
        
        If a cluster has fewer points than the threshold, it is demoted to an outlier cluster.

        Only demotes a single cluster at a time.

        :return: True if a cluster was demoted, False otherwise
        """

        for cluster in self.clusters:
            if cluster.number_of_points < self.max_outlier_points:
                self.outlier_clusters.append(cluster)
                self.clusters.remove(cluster)
                if self.verbose:
                    print(f"Cluster {cluster.id} was demoted to an outlier cluster. Now {len(self.clusters)} regular clusters and {len(self.outlier_clusters)} outlier clusters.")
                return True
        return False
    
    def try_merge_clusters(self, id: int | None, merge_threshold: float) -> tuple[bool, tuple[int, int] | None]:
        """ Try to merge clusters
        
        If two clusters are closer than the threshold, they are merged into a single cluster.
        
        Only merges a single pair of clusters at a time.
        
        :param id: Id of the cluster to be tested for merging
        :param merge_threshold: Threshold for merging clusters
        :return: Tuple of a boolean indicating whether a merge was performed and the ids of the merged clusters
        """

        # Check outlier clusters
        for i, cluster1 in enumerate(self.outlier_clusters):
            for j, cluster2 in enumerate(self.outlier_clusters):
                if id is not None:
                    if cluster1.id != id and cluster2.id != id:
                        continue
                if i != j and cluster1.calculate_distance(cluster2.prototype) < merge_threshold:
                    merge_result = cluster1.try_merge(cluster2)
                    if merge_result is False:
                        continue
                    self.outlier_clusters.remove(cluster2)
                    if self.verbose:
                        print(f"Outlier clusters {cluster1.id} and {cluster2.id} were merged. Now {len(self.clusters)} regular clusters and {len(self.outlier_clusters)} outlier clusters.")
                    return True, (cluster1.id, cluster2.id)
        
    
        # Check merging outliers to regular clusters
        for i, cluster1 in enumerate(self.clusters):
            for j, cluster2 in enumerate(self.outlier_clusters):
                if id is not None:
                    if cluster1.id != id and cluster2.id != id:
                        continue
                if cluster1.calculate_distance(cluster2.prototype) < merge_threshold:
                    merge_result = cluster1.try_merge(cluster2)
                    if merge_result is False:
                        continue
                    self.outlier_clusters.remove(cluster2)
                    if self.verbose:
                        print(f"Outlier cluster {cluster2.id} was merged into regular cluster {cluster1.id}. Now {len(self.clusters)} regular clusters and {len(self.outlier_clusters)} outlier clusters.")
                    return True, (cluster1.id, cluster2.id)
                
        # Check regular clusters
        for i, cluster1 in enumerate(self.clusters):
            for j, cluster2 in enumerate(self.clusters):
                if id is not None:
                    if cluster1.id != id and cluster2.id != id:
                        continue
                if i != j and cluster1.calculate_distance(cluster2.prototype) < merge_threshold:
                    merge_result = cluster1.try_merge(cluster2)
                    if merge_result is False:
                        continue
                    self.clusters.remove(cluster2)
                    if self.verbose:
                        print(f"Clusters {cluster1.id} and {cluster2.id} were merged. Now {len(self.clusters)} regular clusters and {len(self.outlier_clusters)} outlier clusters.")
                    return True, (cluster1.id, cluster2.id)
                
        return False, None


    def fit_point(self, data, verbose: bool = True) -> int:
        """ Fit a new data point to the clusters

        The data point is fitted to the existing clusters. If the distance to the closest cluster is less than `threshold`,
        the data point is added to the cluster. If the distance is greater than `threshold`, a new cluster is created.

        When adding the data point to the cluster, if the new prototype would be less than `min_new_length` times the
        length of the old prototype, the data point is not added to the cluster. Instead, the next closest cluster is
        evaluated, up to the point where the distance is greater than `threshold`. If no cluster is found, a new cluster is
        generated

        :param data: Data point to be fitted to the clusters
        :return cluster index: Index of the cluster to which the data point was added
        """

        if self.threshold is None:
            raise ValueError("Threshold must be set before fitting data points. Use the estimate_threshold method to estimate the threshold.")

        self.max_outlier_points = (np.sum([cluster.number_of_points for cluster in self.clusters]) + np.sum([cluster.number_of_points for cluster in self.outlier_clusters]))*self.outlier_percentage
        if len(self.clusters) == 0:
            self.clusters.append(Cluster(sample=data, alpha=self.alpha, id=next(self.cluster_generator)))
            return self.clusters[-1].id
        else:
            # Try to add to existing clusters
            distances = self.calculate_distances(data)
            while np.min(distances) < self.threshold:
                closest_cluster = np.argmin(distances)
                if distances[closest_cluster] < self.threshold:
                    add_sample_result: bool = self.clusters[closest_cluster].try_add_sample(data,
                                                                                            self.min_new_length)
                    if add_sample_result is True:
                        return self.clusters[closest_cluster].id
                    warnings.warn("Sample was not added to the cluster, trying next closest cluster",
                                  category=UserWarning)
                    distances[closest_cluster] = np.inf
            #Try to add to outlier clusters
            if len(self.outlier_clusters) > 0:
                self.try_demote_cluster()
                distances = np.array([cluster.calculate_distance(data) for cluster in self.outlier_clusters])
                while np.min(distances) < self.threshold:
                    # Try to add to existing outlier clusters
                    closest_cluster = np.argmin(distances)
                    if distances[closest_cluster] < self.threshold:
                        add_sample_result: bool = self.outlier_clusters[closest_cluster].try_add_sample(data,
                                                                                                self.min_new_length)
                        if add_sample_result is True:
                            # Perhaps the outlier cluster has enough points to be promoted
                            if self.try_promote_outlier(self.outlier_clusters[closest_cluster].id):
                                # The outlier cluster was promoted and appended to the regular clusters
                                return self.clusters[-1].id
                            
                            return self.outlier_clusters[closest_cluster].id
                        # Unsuccessful add, try next closest outlier cluster
                        warnings.warn("Sample was not added to the outlier cluster, trying next closest outlier cluster",
                                    category=UserWarning)
                        distances[closest_cluster] = np.inf
                else:
                    self.outlier_clusters.append(Cluster(sample=data, alpha=self.alpha, id=next(self.cluster_generator)))
                    return self.outlier_clusters[-1].id
            else:                # Create first outlier cluster
                self.outlier_clusters.append(Cluster(sample=data, alpha=self.alpha, id=next(self.cluster_generator)))
                return self.outlier_clusters[-1].id

    def calculate_distances(self, sample: np.ndarray) -> np.ndarray:
        """ Calculate the distance of the sample to each cluster

        :param sample: Sample to be compared to the clusters
        :return: Array of distances to all the clusters
        """
        return np.array([cluster.calculate_distance(sample) for cluster in self.clusters])

    def get_prototypes(self) -> list[np.ndarray]:
        """ Get the prototypes of all the clusters

        :return: List of prototypes of all the clusters
        """
        return [cluster.prototype for cluster in self.clusters]
    
    def get_n_clusters(self)->int:
        """ Returns the number of established clusters in the model
        
        :return: Number of established clusters in the model
        """
        return len(self.clusters)

    def get_n_outliers(self)->int:
        """ Returns the number of outlier clusters in the model
        
        :return: Number of outlier clusters in the model
        """
        return len(self.outlier_clusters)


import numpy as np
from collections.abc import Generator, Iterable

class Segmentor:
    
    dx_threshold: float
    min_segment_length: int
    max_segment_length: int
    def __init__(self, dx_threshold: float, min_segment_length: int = 10, max_segment_length: int = np.inf):
        self.dx_threshold = dx_threshold
        if min_segment_length < 1:
            raise ValueError("min_segment_length must be greater than 0")
        self.min_segment_length = min_segment_length
        if max_segment_length < min_segment_length:
            raise ValueError("max_segment_length must be greater than min_segment_length")
        self.max_segment_length = max_segment_length

    def get_segments_with_indices(self, data: Iterable) -> Generator[list[tuple[np.ndarray, np.ndarray]], None, None]:
        
        segment: list[float] = []
        for i,x in enumerate(data):
            if len(segment) < self.min_segment_length or len(segment) == 0:
                segment.append(x)
                continue

            dx: float = x - segment[-1]
            if dx < -self.dx_threshold or len(segment) >= self.max_segment_length:
                yield np.array(segment), np.arange(start=i - len(segment), stop=i, step=1)
                segment = [segment[-1]]

            segment.append(x)

        return None
    
    def get_segments(self, data: Iterable) -> Generator[list[np.ndarray], None, None]:
        for segment, _ in self.get_segments_with_indices(data):
            yield segment


import requests
from io import BytesIO
from zipfile import ZipFile
import os
import datetime
import pickle as pkl
import numpy as np


def download_industrial_dataset(data_folder: str, make_dir: bool = True, skip_if_exists: bool = True) -> None:
    """
    Helper function to dowload the industrial dataset to the specified folder. If the folder is empty, the dataset will
    not be downloaded.

    Upon download, the dataset will be unzipped to the specified folder.
    :param data_folder: Target folder for the dataset
    :param make_dir: Should directory be created if it does not exist?
    :param skip_if_exists: Should skip download if the folder is not empty?
    :return: None
    """
    # URL taken from https://data.mendeley.com/datasets/ypzswhhzh9/2 - see link under "Download All" button.
    dataset_url = "https://prod-dcd-datasets-cache-zipfiles.s3.eu-west-1.amazonaws.com/ypzswhhzh9-2.zip"

    if not os.path.exists(data_folder):
        if not make_dir:
            raise FileNotFoundError("Dataset folder does not exist")
        os.makedirs(data_folder)

    if not os.path.isdir(data_folder):
        raise ValueError("Path supplied is not a directory")

    if len(os.listdir(data_folder)) > 0:
        if skip_if_exists:
            print("Supplied folder is not empty. Skipping dataset download")
            return
        raise FileExistsError("Supplied folder is not empty")
    print("Starting download of dataset")
    tic = datetime.datetime.now()
    response = requests.get(dataset_url)
    print(f"Download complete, took {(datetime.datetime.now() - tic).total_seconds():.1f}s")
    ZipFile(BytesIO(response.content)).extractall(data_folder)
    print(f"Dataset unzipped to {os.path.abspath(data_folder)}")


def load_industrial_dataset(folder: str, download: bool = True, make_dir: bool = True) -> tuple[dict, dict]:
    """ Helper function to load the industrial dataset as a tuple of two dictionaries, one for pressure and one for
    current. Each dictionary contains data and labels keys. Data is a list of numpy arrays, each representing a
    segment of the signal. Labels is a list of integers, each representing a class of the segment.

    :param folder: The folder where the dataset is stored
    :param download: Should try to download the dataset?
    :param make_dir: Should create the directory if it does not exist?
    :return: A tuple of two dictionaries, one for pressure and one for current.
    """
    if download:
        download_industrial_dataset(folder, make_dir=make_dir, skip_if_exists=True)

    suffix: str = os.path.join("Pneumatic Pressure and Electrical Current Time Series in Manufacturing", "Pickle")

    with open(os.path.join(folder, suffix, 'segmented_pressure.pkl'), 'rb') as f:
        pressure = pkl.load(f)

    with open(os.path.join(folder, suffix, "segmented_current.pkl"), 'rb') as f:
        current = pkl.load(f)

    return pressure, current


def get_industrial_data_grouped_by_class(download: bool = True,
                                         folder: str = "Data",
                                         min_samples_per_class: int = 5,
                                         min_sample_length: int = 11,
                                         max_sample_length: int = 600) -> tuple[
    list[list[np.ndarray]], list[list[np.ndarray]]]:
    """ Helper function to load the industrial dataset and group the segments by class. The function will return two
    lists of lists (one for pressure, one for current), where each inner list contains segments of the same class.

    :param download: Should try to download the dataset?
    :param folder: The folder where the dataset is stored
    :param min_samples_per_class: Minimum number of samples per class. Classes with fewer samples will be discarded.
    :param min_sample_length: Minimum length of a sample. Shorter samples will be discarded.
    :param max_sample_length: Maximum length of a sample. Longer samples will be discarded.
    :return: Two lists of lists, where each inner list contains segments of the same class. First list is for pressure,
    second for current.
    """

    pressure, current = load_industrial_dataset(folder=folder, download=download)
    datasets_dict: dict[int, list[np.ndarray]] = {
        state: [series.to_numpy() for i, series in enumerate(pressure['data']) if
                pressure['labels'][i] == state and len(series) >= min_sample_length and len(
                    series) <= max_sample_length] for state in np.unique(pressure['labels'])}
    datasets_dict = {k: v for k, v in datasets_dict.items() if min_samples_per_class < len(v)}
    datasets_pressure = [v for k, v in datasets_dict.items()]

    datasets_dict: dict[int, list[np.ndarray]] = {
        state: [series.to_numpy() for i, series in enumerate(current['data']) if
                current['labels'][i] == state and len(series) >= min_sample_length and len(series) <= max_sample_length]
        for state in np.unique(current['labels'])}
    datasets_dict = {k: v for k, v in datasets_dict.items() if min_samples_per_class < len(v)}
    datasets_current = [v for k, v in datasets_dict.items()]

    return datasets_pressure, datasets_current


def get_industrial_unlabeled(download: bool = True,
                             folder: str = "Data",
                             min_sample_length: int = 11,
                             max_sample_length: int = 600) -> tuple[list[np.ndarray], list[np.ndarray]]:
    """ Helper function to load the industrial dataset and group the segments by class. The function will return two
    lists (one for pressure, one for current)

    :param download: Should try to download the dataset?
    :param folder: The folder where the dataset is stored
    :param min_samples_per_class: Minimum number of samples per class. Classes with fewer samples will be discarded.
    :param min_sample_length: Minimum length of a sample. Shorter samples will be discarded.
    :param max_sample_length: Maximum length of a sample. Longer samples will be discarded.
    :return: Two lists, one for pressure, second for current.
    """

    pressure, current = load_industrial_dataset(folder=folder, download=download)

    pressure_return = [x.values for x in pressure['data'] if
                       len(x) >= min_sample_length and len(x) <= max_sample_length]
    current_return = [x.values for x in current['data'] if len(x) >= min_sample_length and len(x) <= max_sample_length]

    return pressure_return, current_return

def get_industrial_labeled(download: bool = True,
                             folder: str = "Data",
                             min_sample_length: int = 11,
                             max_sample_length: int = 600,
                             min_samples_per_class: int = 1) -> tuple[list[np.ndarray], list[np.ndarray], list[int]]:
    """ Helper function to load the industrial dataset and group the segments by class. The function will return two
    lists (one for pressure, one for current)

    :param download: Should try to download the dataset?
    :param folder: The folder where the dataset is stored
    :param min_samples_per_class: Minimum number of samples per class. Classes with fewer samples will be discarded.
    :param min_sample_length: Minimum length of a sample. Shorter samples will be discarded.
    :param max_sample_length: Maximum length of a sample. Longer samples will be discarded.
    :param min_samples_per_class: Minimum number of samples per class. Classes with fewer samples will be discarded.
    :return: Two lists, one for pressure, second for current.
    """

    pressure, current = load_industrial_dataset(folder=folder, download=download)

    pressure_return = [x.values for x in pressure['data'] if
                       len(x) >= min_sample_length and len(x) <= max_sample_length]
    current_return = [x.values for x in current['data'] if len(x) >= min_sample_length and len(x) <= max_sample_length]

    labels = [x for i, x in enumerate(pressure['labels']) if len(pressure['data'][i]) >= min_sample_length and len(pressure['data'][i]) <= max_sample_length]

    if min_samples_per_class <= 1:
        return pressure_return, current_return, labels
    
    class_sizes = [labels.count(x) for x in np.unique(labels)]
    excluded_classes = [x for i, x in enumerate(np.unique(labels)) if class_sizes[i] < min_samples_per_class]

    pressure_return = [x for i, x in enumerate(pressure_return) if labels[i] not in excluded_classes]
    current_return = [x for i, x in enumerate(current_return) if labels[i] not in excluded_classes]
    labels = [x for x in labels if x not in excluded_classes]   

    return pressure_return, current_return, labels


import numpy as np
from tslearn.metrics import cdist_normalized_cc, y_shifted_sbd_vec

def sbd(X: list[np.ndarray], n_iter: int = 10) -> np.ndarray:
    """ Shape-Based-Distance (SBD) barycenter computation
    
    Implementation inspired by the tslearn library.
    See https://github.com/tslearn-team/tslearn/blob/5568c026db4b4380b99095827e0573a8f55a81f0/tslearn/clustering/kshape.py#L118

    SBD barycenter computation is first described in [1] as part of the k-Shape algorithm.



    [1] J. Paparrizos & L. Gravano. k-Shape: Efficient and Accurate
        Clustering of Time Series. SIGMOD 2015. pp. 1855-1870.

    """
    p = X[0].reshape(-1, 1)
    
    # In [1], the algorithm is used as part of the k-Shape algorithm. There it is
    # executed a number of times, as part of the Refinement step.
    # Here, we run the algorithm a number of times, to enable prototype convergence.
    for i in range(n_iter):

        sz = len(X[0])
        for x in X: assert x.shape[0] == sz
        X = [x.reshape(-1, 1) for x in X]
        X_np = np.array(X)
        assert X_np.shape[0] == len(X)
        assert X_np.shape[1] == sz
        assert X_np.shape[2] == 1
        X = X_np

        norms = np.linalg.norm(X, axis=(1, 2))

        sz = X.shape[1]
        Xp = y_shifted_sbd_vec(p, X,
                                norm_ref=-1,
                                norms_dataset=norms)
        S = np.dot(Xp[:, :, 0].T, Xp[:, :, 0])
        Q = np.eye(sz) - np.ones((sz, sz)) / sz
        M = np.dot(Q.T, np.dot(S, Q))
        _, vec = np.linalg.eigh(M)
        mu_k = vec[:, -1].reshape((sz, 1))

        # The way the optimization problem is (ill-)formulated, both mu_k and
        # -mu_k are candidates for barycenters
        # In the following, we check which one is best candidate
        dist_plus_mu = np.sum(np.linalg.norm(Xp - mu_k, axis=(1, 2)))
        dist_minus_mu = np.sum(np.linalg.norm(Xp + mu_k, axis=(1, 2)))
        if dist_minus_mu < dist_plus_mu:
            mu_k *= -1
        p = mu_k

    return p


import warnings
import numpy as np
from eral.eral_score import eral_score_with_normalization
from eral.time_series_helpers import _remove_nan_padding_single
from eral.time_series_alignment import sync_n_series_to_prototype, get_score_and_shift

class Cluster:
    """ Class representing a cluster in the evolving time series clustering

    The cluster is represented by two structures:
    - PSV (Prototype Shape Vector), which carries the shape of the prototype, representing the shapes of all samples used in
        updating the prototype
    - PCV (Prototype Confidence Vector), which describes how many input samples were used in conditioning the prototype at each
        time point

    The final prototype is obtained by combining PSV and PCV. The prototype is
    only defined in the time points where the PCV is greater than some user-defined threshold, for
    example 0.8 indicating that the prototype is defined at indices where at least 80% of the samples were used in
    conditioning the prototype.

    """

    psv: np.ndarray
    pcv: np.ndarray
    number_of_points: int
    alpha: float
    id: int
    eral_exclusion_zone: float = 0.2

    def __init__(self, sample: np.ndarray, id: int,  alpha: float = 0.1, exclusion_zone: float = 0.2):
        """ Initialize the cluster with a sample.

        Choosing :param:`alpha`: 1.0 --> only use region where all samples were used, results in a short prototype, 
        0.0 --> use all samples, results in a long prototype

        :param sample: Sample to be used as the prototype
        :param alpha: Threshold for the PCV: required proportion of samples to be used in conditioning the prototype.
        :param exclusion_zone: Exclusion zone for the alignment of the new sample to the prototype
        """

        self.psv = sample
        self.pcv = np.ones_like(sample)
        self.number_of_points = 1
        self.alpha = alpha
        self.id = id
        self.eral_exclusion_zone = exclusion_zone

    def try_add_sample(self, sample: np.ndarray, min_new_length: float = 0.8) -> bool:
        """ Try to add a new sample to the cluster

        The new sample is added to the cluster, if the new prototype would not be less than `min_new_length` times
        the length of the old prototype. If the new prototype would be too short, the sample is not added.

        If the sample is added, the prototype is updated, and the number of points is increased. True is returned.
        If the sample is not added, False is returned.

        :param sample: New sample to be added to the cluster
        :param min_new_length: Minimum length of the new prototype as a proportion of the old prototype
        :return: True if the sample was added, False if the sample was not added
        """

        if self.number_of_points == 0:
            raise ValueError("Cluster is not initialized, can not add samples")

        if min_new_length > 1:
            raise ValueError("min_new_length should be less than 1")
        
        new_psv, new_pcv, common_time = self._get_new_components(sample)

        start_idx, end_idx = self._calculate_crisp_prototype_boundaries(new_psv, new_pcv, self.alpha)
        new_prototype = new_psv[start_idx:end_idx]

        if len(new_prototype) / len(self.prototype) < min_new_length:
            return False

        self.psv = new_psv
        self.pcv = new_pcv
        self.number_of_points += 1
        # self.common_time = common_time

        return True

    def _get_alignment_of_sample_to_psv(self, sample: np.ndarray) -> int:
        """ Get the lag to be applied to the sample to align it to the PSV

        Using the crisp prototype, the lag is calculated. Then this lag is modified so that it can be applied to align the sample and the
        PSV.

        The sample must not be simply applied to the PSV, since the PSV contains
        ill-defined values at the edges.

        Attention: "Apply zero mean" is set to False in the call to :func:`eral.time_series_alignment.get_score_and_shift`.

        :param sample: Sample to be aligned to the prototype
        :return: Lag suitable to be applied to the sample to align it to the PSV
        """

        # Get current crisp prototype
        start_idx, end_idx = self._get_crisp_prototype_boundaries()
        prototype = self.psv[start_idx:end_idx]

        # Get optimal alignment of the new sample to the crisp prototype
        alignment: int = get_score_and_shift(prototype, sample, exclusion_zone=self.eral_exclusion_zone, apply_zero_mean=False)[1] # TODO: apply zero mean is fixed, check if it should be changed

        return alignment + start_idx

    def _get_shifted_sample_and_psv_with_alignment(self, sample: np.ndarray) -> tuple[np.ndarray, np.ndarray, int, np.ndarray]:
        """ Shift the sample and the PSV to align them

        Only the crisp prototype is used in the alignment. The PSV is then shifted to match the
        alignment of the sample and the crisp prototype.

        :param sample: Sample to be aligned to the prototype
        :return: Tuple of the shifted sample, the shifted PSV, alignment
        """

        alignment: int = self._get_alignment_of_sample_to_psv(sample)

        # Use obtained optimal alignment to shift the PSV
        psv = self.psv
        common_time, shifted_sample, shifted_psv = sync_n_series_to_prototype(prototype=psv,
                                                                              series=[sample],
                                                                              shifts=[alignment],
                                                                              exclusion_zone=self.eral_exclusion_zone)
        
        return shifted_sample[0], shifted_psv, alignment, common_time

    def _get_shifted_sample_and_psv(self, sample: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        """ Shift the sample and the PSV to align them 
        
        Only the crisp prototype is used in the alignment. The PSV is then shifted to match the
        alignment of the sample and the crisp prototype.

        :param sample: Sample to be aligned to the prototype
        :return: Tuple of the shifted sample and the shifted PSV
        """

        shifted_sample, shifted_psv, _, _ = self._get_shifted_sample_and_psv_with_alignment(sample)
        
        return shifted_sample, shifted_psv


    def _generate_new_pcv(self, start_idx: int, end_idx: int) -> np.ndarray:
        """ Generate a new PCV when adding new sample to the cluster 
        
        The region where the new sample is added to the cluster has increased support after the addition. The
        PCV is updated to reflect this change.

        :param start_idx: Start index of the new sample in the PSV
        :param end_idx: End index of the new sample in the PSV
        :return: New PCV
        """

        previous_number_of_points = self.number_of_points
        new_number_of_points = previous_number_of_points + 1

        old_pcv = self.pcv
        new_pcv = old_pcv.copy() * previous_number_of_points
        if start_idx < 0:
            new_pcv = np.concatenate([np.zeros(-start_idx), new_pcv])
            end_idx+=-start_idx
            start_idx = 0
        if end_idx > len(new_pcv):
            new_pcv = np.concatenate([new_pcv, np.zeros(end_idx - len(new_pcv))])
            end_idx = len(new_pcv)

        new_pcv[start_idx:end_idx] += 1

        new_pcv /= new_number_of_points

        return new_pcv

    def _generate_new_psv(self, aligned_sample: np.ndarray, aligned_psv: np.ndarray) -> np.ndarray:
        """ Generate a new PSV by adding the new aligned sample to the existing PSV """

        assert len(aligned_sample) == len(aligned_psv), ("Aligned sample and aligned PSV should have the same length. "
                                                                        f"Lengths are {len(aligned_sample)} and "
                                                                        f"{len(aligned_psv)}")
        
        # Calculate the weights
        weight_new_sample = 1 / (self.number_of_points + 1)
        weight_old_prototype = self.number_of_points / (self.number_of_points + 1)

        # Calculate the new prototype
        
        aligned_psv_nan_mask = np.isnan(aligned_psv)
        aligned_sample_nan_mask = np.isnan(aligned_sample)

        sample_weights_vector = np.zeros_like(aligned_sample)+weight_new_sample
        prototype_weights_vector = np.zeros_like(aligned_psv)+weight_old_prototype
        # Weights should be weight_new_sample and weight_old_prototype where neither of the values is NaN
        pass

        # At regions where only the sample is NaN, the prototype should be used
        sample_weights_vector[aligned_sample_nan_mask] = 0
        prototype_weights_vector[aligned_sample_nan_mask] = 1

        # At regions where only the prototype is NaN, the sample should be used
        sample_weights_vector[aligned_psv_nan_mask] = 1
        prototype_weights_vector[aligned_psv_nan_mask] = 0

        # The sum of the weights should be 1
        assert np.allclose(sample_weights_vector+prototype_weights_vector, 1), ("Sample and prototype weights should sum to 1")

        weighted_new_sample = aligned_sample * sample_weights_vector
        weighted_old_prototype = aligned_psv * prototype_weights_vector

        new_prototype = np.nansum(np.array([weighted_old_prototype, weighted_new_sample]), axis=0)
        
        return new_prototype

    def _get_new_components(self, sample: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """ Get the new PSV, the new PCV, and the common time points
        
        The new sample is aligned to the existing crisp prototype. The lag is used to align the sample and the
        PSV. The PSV is then updated by adding the new sample to the prototype.

        The PCV is updated to reflect the latest update.

        This method does not update the cluster, only calculates the new components.

        :param sample: New sample to be added to the cluster
        :return: Tuple of the new PSV, the new PCV, and the common time points
        """

        shifted_sample, shifted_psv, alignment, common_time = self._get_shifted_sample_and_psv_with_alignment(sample)
        assert len(shifted_sample) == len(shifted_psv), ("Shifted sample and shifted PSV should have the same length. "
                                                                        f"Lengths are {len(shifted_sample)} and "
                                                                        f"{len(shifted_psv)}")

        new_pcv = self._generate_new_pcv(start_idx=alignment, end_idx=alignment+len(sample))

        new_psv = self._generate_new_psv(shifted_sample, shifted_psv)

        assert len(new_psv) == len(new_pcv), ("New PSV and new PCV should have the same length. "
                                                                            f"Lengths are {len(new_psv)} and "
                                                                            f"{len(new_pcv)}")
        
        return new_psv, new_pcv, common_time

    def add_sample(self, sample: np.ndarray):
        """ Add a new sample to the cluster

        The new sample is aligned to the existing crisp prototype. The lag is used to align the sample and the
        PSV. The PSV is then updated by adding the new sample to the prototype.

        The PCV is updated to reflect the latest update.

        :param sample: New sample to be added to the cluster
        :return: None
        """

        new_psv, new_pcv, common_time = self._get_new_components(sample)

        # self.common_time = common_time
        self.psv = new_psv
        self.pcv = new_pcv
        self.number_of_points += 1

    @staticmethod
    def distance(prototype: np.ndarray, sample: np.ndarray, eral_exclusion_zone: float) -> float:
        """ Calculate the distance between the prototype and the sample

        ERAL score with normalization is used to calculate the error in alignment of the sample to the prototype.
        Normalization in ERAL score is done by the length of the overlap of the sample and the prototype.

        ERAL score is further multiplied by (len_sample*len_prototype)/(overlap_len^2) to penalize short overlaps.

        :param prototype: Prototype to be compared to the sample
        :param sample: Sample to be compared to the prototype
        :param eral_exclusion_zone: Exclusion zone for the alignment of the new sample to the prototype
        :return: Distance
        """

        a = _remove_nan_padding_single(prototype)
        b = _remove_nan_padding_single(sample)
        eral_score, normalization = eral_score_with_normalization(a, b)

        if eral_exclusion_zone > 0:
            # Exclude the forbidden shifts
            exclusion_zone = int(len(eral_score) * eral_exclusion_zone)
            eral_score[:exclusion_zone] = np.inf
            eral_score[-exclusion_zone:] = np.inf

        eral_dist = np.min(eral_score)
        normalization = normalization[np.argmin(eral_score)]

        overlap_len = normalization ** 2

        prototype_len = len(_remove_nan_padding_single(prototype))
        sample_len = len(_remove_nan_padding_single(sample))
        factor = (prototype_len * sample_len) / (overlap_len ** 2)

        return eral_dist * normalization * (factor ** 2)

    def calculate_distance(self, sample: np.ndarray) -> float:
        """ Calculate the distance of the sample to the prototype, using the ERAL score

        ERAL score with normalization is used to calculate the error in alignment of the sample to the prototype.
        Normalization in ERAL score is done by the length of the overlap of the sample and the prototype.

        ERAL score is further multiplied by (len_sample*len_prototype)/(overlap_len^2) to penalize short overlaps.

        :param sample: Sample to be compared to the prototype
        :return: Distance
        """

        return self.distance(self.prototype, sample, self.eral_exclusion_zone)


    @staticmethod
    def _calculate_crisp_prototype_boundaries(psv: np.ndarray, pcv: np.ndarray, alpha: float) -> tuple[int, int]:
        """ Get the start and end index of the crisp prototype in the PSV

        The crisp prototype is defined in the time points where the PCV is greater than some user-defined
        threshold, for example 0.8 indicating that the prototype is defined at indices where at least 80% of the
        samples were used in conditioning the prototype.

        Choosing :param:`alpha`: 1.0 --> only use region where all samples were used, results in a short prototype, 
        0.0 --> use all samples, results in a long prototype

        :param psv: Prototype Shape Vector
        :param pcv: Prototype Confidence Vector
        :param alpha: Threshold for the PCV: required proportion of samples to be used in conditioning the prototype.
        :return: Tuple of the start and end index of the crisp prototype in the PSV
        """

        assert len(psv) == len(pcv), ("PSV and PCV should have the same length")

        bounding_filter = pcv >= alpha
        start_idx = np.argmax(bounding_filter)
        end_idx = len(bounding_filter) - np.argmax(bounding_filter[::-1])
        return start_idx, end_idx


    def _get_crisp_prototype_boundaries(self) -> tuple[int, int]:
        """ Get the start and end index of the crisp prototype in the PSV

        The crisp prototype is defined in the time points where the PCV is greater than some user-defined
        threshold, for example 0.8 indicating that the prototype is defined at indices where at least 80% of the
        samples were used in conditioning the prototype.

        :return: Tuple of the start and end index of the crisp prototype in the PSV
        """

        start_idx, end_idx = self._calculate_crisp_prototype_boundaries(self.psv, self.pcv, self.alpha)

        return start_idx, end_idx


    def _prototype_defuzzification(self) -> np.ndarray:
        """ Returns the prototype of the cluster

        The prototype is obtained by combining the PSV and the PCV. The prototype is
        only defined in the time points where the PCV is greater than some user-defined threshold, for
        example 0.8 indicating that the prototype is defined at indices where at least 80% of the samples were used
        in conditioning the prototype.

        :return: Crisp prototype
        """

        start_idx, end_idx = self._get_crisp_prototype_boundaries()

        return self.psv[start_idx:end_idx]

    @property
    def prototype(self) -> np.ndarray:
        """ Get the prototype of the cluster """
        return self._prototype_defuzzification()


    def copy(self, deep: bool = False, new_id: int | None = None) -> 'Cluster':
        """ Create a copy of the cluster

        :param deep: If True, a deep copy is made. If False, a shallow copy is made.
        :return: Copy of the cluster
        """

        if new_id is None:
            new_id = self.id

        if not deep:
            new_cluster = self.__class__(sample=self.psv)
            new_cluster.pcv = self.pcv.copy()
            new_cluster.number_of_points = self.number_of_points
            new_cluster.alpha = self.alpha
            new_cluster.id = new_id
            new_cluster.eral_exclusion_zone = self.eral_exclusion_zone
            return new_cluster
        
        new_cluster = self.__class__(sample=self.psv.copy(), id = new_id)
        new_cluster.pcv = self.pcv.copy()
        new_cluster.number_of_points = self.number_of_points
        new_cluster.alpha = self.alpha
        new_cluster.eral_exclusion_zone = self.eral_exclusion_zone
        return new_cluster
    
    def merge(self, other_cluster: 'Cluster', new_alpha: float | None = None, new_id: int | None = None):
        """ Force merge the other cluster into this cluster
        
        The other cluster is added to this cluster. The prototype is updated by adding the prototype of the other
        cluster to the prototype of this cluster. The PCV is updated to reflect the new number of points.
        
        :param other_cluster: Cluster to be merged into this cluster
        :param new_alpha: New alpha value for the merged cluster. If None, the alpha value of this cluster is used.
        :param new_id: New id for the merged cluster. If None, the id of this cluster is used.
        """

        success = self.try_merge(other_cluster, new_alpha, new_id, new_min_length=0)
        if not success:
            raise Exception("Merge was not successful")
        return

    def try_merge(self, other_cluster: 'Cluster', new_alpha: float | None = None, new_id: int | None = None, min_new_length: float = 0.9) -> bool:
        """ Try to merge the other cluster into this cluster

        The other cluster is added to this cluster. The prototype is updated by adding the prototype of the other
        cluster to the prototype of this cluster. The PCV is updated to reflect the new number of points.

        If the new prototype would be less than `min_new_length` times the length of the old prototype, the merge is not
        performed, False is returned.

        :param other_cluster: Cluster to be merged into this cluster
        :param new_alpha: New alpha value for the merged cluster. If None, the alpha value of this cluster is used.
        :param new_id: New id for the merged cluster. If None, the id of this cluster is used.
        :param min_new_length: Minimum length of the new prototype as a proportion of the old prototype.
        :return: True if the merge was successful, False if the merge was not successful (new prototype would be too short)
        """

        if not isinstance(other_cluster, Cluster):
            raise ValueError(f"Can only merge with another Cluster object, not {type(other_cluster)}")

        my_psv = self.psv
        my_pcv = self.pcv
        my_count = self.number_of_points
        other_psv = other_cluster.psv
        other_pcv = other_cluster.pcv
        other_count = other_cluster.number_of_points

        if new_alpha is None:
            new_alpha = self.alpha
        if new_id is None:
            new_id = self.id

        # my_prototype = self.prototype
        # other_prototype = other_cluster.prototype
        my_prototype_start, my_prototype_end = self._get_crisp_prototype_boundaries()
        my_prototype = my_psv[my_prototype_start:my_prototype_end]
        other_prototype_start, other_prototype_end = other_cluster._get_crisp_prototype_boundaries()
        other_prototype = other_psv[other_prototype_start:other_prototype_end]
        
        # Align the prototypes
        prototype_alignment = get_score_and_shift(my_prototype, other_prototype, exclusion_zone=self.eral_exclusion_zone, apply_zero_mean=False)[1] # TODO: apply zero mean is fixed, check if it should be changed
        alignment = prototype_alignment + my_prototype_start - other_prototype_start
        

        # Calculate the new PCV
        # alignment: int = self._get_alignment_of_sample_to_psv(other_prototype)

        _, shifted_other_pcv, shifted_my_pcv = sync_n_series_to_prototype(prototype=my_pcv,
                                                                                    series=[other_pcv],
                                                                                    shifts=[alignment])
        shifted_other_pcv = shifted_other_pcv[0]
        shifted_other_pcv = np.nan_to_num(shifted_other_pcv, nan=0)
        shifted_my_pcv = np.nan_to_num(shifted_my_pcv, nan=0)
        new_pcv = (my_count * shifted_my_pcv + other_count * shifted_other_pcv) / (my_count + other_count)
        
        # Calculate the new PSV
        _, shifted_other_psv, shifted_my_psv = sync_n_series_to_prototype(prototype=my_psv,
                                                                                    series=[other_psv],
                                                                                    shifts=[alignment])
        shifted_other_psv = shifted_other_psv[0]

        weight_my_psv = my_count / (my_count + other_count)
        weight_other_psv = other_count / (my_count + other_count)

        aligned_my_psv_nan_mask = np.isnan(shifted_my_psv)
        aligned_other_psv_nan_mask = np.isnan(shifted_other_psv)
        
        my_psv_weights_vector = np.zeros_like(shifted_my_psv)+weight_my_psv
        other_psv_weights_vector = np.zeros_like(shifted_other_psv)+weight_other_psv        
        # Weights should be weight_my_psv and weight_other_psv where neither of the values is NaN
        pass

        # At regions where only my_psv is NaN, other_psv should be used
        my_psv_weights_vector[aligned_my_psv_nan_mask] = 0
        other_psv_weights_vector[aligned_my_psv_nan_mask] = 1

        # At regions where only other_psv is NaN, my_psv should be used
        my_psv_weights_vector[aligned_other_psv_nan_mask] = 1
        other_psv_weights_vector[aligned_other_psv_nan_mask] = 0

        # The sum of the weights should be 1
        assert np.allclose(my_psv_weights_vector+other_psv_weights_vector, 1), ("My PSV and other PSV weights should sum to 1")

        weighted_my_psv = shifted_my_psv * my_psv_weights_vector
        weighted_other_psv = shifted_other_psv * other_psv_weights_vector

        new_psv = np.nansum(np.array([weighted_my_psv, weighted_other_psv]), axis=0)

        # Remove padding where pcv is 0
        start_idx = np.argmax(new_pcv>0)
        end_idx = len(new_pcv) - np.argmax(new_pcv[::-1]>0)
        new_psv = new_psv[start_idx:end_idx]
        new_pcv = new_pcv[start_idx:end_idx]
        
        assert len(new_psv) == len(new_pcv), ("New PSV and new PCV should have the same length.")

        final_prototype_start_idx, final_prototype_end_idx = self._calculate_crisp_prototype_boundaries(new_psv, new_pcv, new_alpha)
        final_prototype_length = final_prototype_end_idx - final_prototype_start_idx

        if final_prototype_length < len(my_prototype)*min_new_length:
            return False

        # Update the cluster
        self.psv = new_psv
        self.pcv = new_pcv
        self.number_of_points += other_count
        self.alpha = new_alpha
        self.id = new_id

        return True
